@@ -2,11 +2,12 @@ import logging
 import os
 from weakref import ref
 
-from ..config import config
-from ..utils import format_bytes, PeriodicCallback, log_errors, ignoring
+import dask
 
 from .adaptive import Adaptive
 
+from ..utils import format_bytes, PeriodicCallback, log_errors, ignoring
+
 logger = logging.getLogger(__name__)
 
 
@@ -14,7 +15,7 @@ class Cluster(object):
     """ Superclass for cluster objects
 
     This expects a local Scheduler defined on the object.  It provides
-    common methods, and an IPython widget display.
+    common methods and an IPython widget display.
 
     Clusters inheriting from this class should provide the following:
 
@@ -30,6 +31,22 @@ class Cluster(object):
     This will provide a general ``scale`` method as well as an IPython widget
     for display.
 
+    Examples
+    --------
+
+    >>> from distributed.deploy import Cluster
+    >>> class MyCluster(cluster):
+    ...     def scale_up(self, n):
+    ...         ''' Bring the total worker count up to n '''
+    ...         pass
+    ...     def scale_down(self, workers):
+    ...         ''' Close the workers with the given addresses '''
+    ...         pass
+
+    >>> cluster = MyCluster()
+    >>> cluster.scale(5)                       # scale manually
+    >>> cluster.adapt(minimum=1, maximum=100)  # scale automatically
+
     See Also
     --------
     LocalCluster: a simple implementation with local workers
@@ -121,7 +138,7 @@ class Cluster(object):
         layout = Layout(width='150px')
 
         if 'bokeh' in self.scheduler.services:
-            template = config.get('diagnostics-link', 'http://{host}:{port}/status')
+            template = dask.config.get('distributed.dashboard.link')
 
             host = self.scheduler.address.split('://')[1].split(':')[0]
             port = self.scheduler.services['bokeh'].port
@@ -163,7 +180,6 @@ class Cluster(object):
         adapt.on_click(adapt_cb)
 
         def scale_cb(b):
-            print('Hello!')
             with log_errors():
                 n = request.value
                 with ignoring(AttributeError):
