@@ -124,16 +124,16 @@ def test_adaptive_local_cluster_multi_workers():
         yield cluster._close()
 
 
-@gen_cluster(client=True, ncores=[('127.0.0.1', 1)] * 10)
+@gen_cluster(client=True, ncores=[('127.0.0.1', 1)] * 10, active_rpc_timeout=10)
 def test_adaptive_scale_down_override(c, s, *workers):
     class TestAdaptive(Adaptive):
         def __init__(self, *args, **kwargs):
             self.min_size = kwargs.pop("min_size", 0)
             Adaptive.__init__(self, *args, **kwargs)
 
-        def workers_to_close(self):
+        def workers_to_close(self, **kwargs):
             num_workers = len(self.scheduler.workers)
-            to_close = self.scheduler.workers_to_close()
+            to_close = self.scheduler.workers_to_close(**kwargs)
             if num_workers - len(to_close) < self.min_size:
                 to_close = to_close[:num_workers - self.min_size]
 
@@ -166,7 +166,7 @@ def test_min_max():
     yield cluster._start()
     try:
         adapt = Adaptive(cluster.scheduler, cluster, minimum=1, maximum=2,
-                         interval='20 ms')
+                         interval='20 ms', wait_count=10)
         c = yield Client(cluster, asynchronous=True, loop=loop)
 
         start = time()
@@ -196,7 +196,7 @@ def test_min_max():
         start = time()
         while len(cluster.scheduler.workers) != 1:
             yield gen.sleep(0.01)
-            assert time() < start + 1
+            assert time() < start + 2
         assert frequencies(pluck(1, adapt.log)) == {'up': 2, 'down': 1}
     finally:
         yield c._close()
@@ -319,3 +319,76 @@ def test_no_more_workers_than_tasks():
     finally:
         yield client._close()
         yield cluster._close()
+
+
+def test_basic_no_loop():
+    try:
+        with LocalCluster(0, scheduler_port=0, silence_logs=False,
+                          diagnostics_port=None) as cluster:
+            with Client(cluster) as client:
+                cluster.adapt()
+                future = client.submit(lambda x: x + 1, 1)
+                assert future.result() == 2
+            loop = cluster.loop
+    finally:
+        loop.add_callback(loop.stop)
+
+
+@gen_test(timeout=None)
+def test_target_duration():
+    """ Ensure that redefining adapt with a lower maximum removes workers """
+    cluster = yield LocalCluster(0, asynchronous=True, processes=False,
+                                 scheduler_port=0, silence_logs=False,
+                                 diagnostics_port=None)
+    client = yield Client(cluster, asynchronous=True)
+    adaptive = cluster.adapt(interval='20ms', minimum=2, target_duration='5s')
+
+    cluster.scheduler.task_duration['slowinc'] = 1
+
+    try:
+        while len(cluster.scheduler.workers) < 2:
+            yield gen.sleep(0.01)
+
+        futures = client.map(slowinc, range(100), delay=0.3)
+
+        while len(adaptive.log) < 2:
+            yield gen.sleep(0.01)
+
+        assert adaptive.log[0][1:] == ('up', {'n': 2})
+        assert adaptive.log[1][1:] == ('up', {'n': 20})
+
+    finally:
+        yield client._close()
+        yield cluster._close()
+
+
+@gen_test(timeout=None)
+def test_worker_keys():
+    """ Ensure that redefining adapt with a lower maximum removes workers """
+    cluster = yield LocalCluster(0, asynchronous=True, processes=False,
+                                 scheduler_port=0, silence_logs=False,
+                                 diagnostics_port=None)
+
+    try:
+        yield [cluster.start_worker(name='a-1'),
+               cluster.start_worker(name='a-2'),
+               cluster.start_worker(name='b-1'),
+               cluster.start_worker(name='b-2')]
+
+        while len(cluster.scheduler.workers) != 4:
+            yield gen.sleep(0.01)
+
+        def key(ws):
+            return ws.name.split('-')[0]
+        cluster._adaptive_options = {'worker_key': key}
+
+        adaptive = cluster.adapt(minimum=1)
+        yield adaptive._adapt()
+
+        while len(cluster.scheduler.workers) == 4:
+            yield gen.sleep(0.01)
+
+        names = {ws.name for ws in cluster.scheduler.workers.values()}
+        assert names == {'a-1', 'a-2'} or names == {'b-1', 'b-2'}
+    finally:
+        yield cluster._close()
